from django.apps import apps
from django.db import models, transaction
from pg_partitioning.signals import post_attach_partition, post_create_partition, post_detach_partition

from .constants import (
    SQL_APPEND_TABLESPACE,
    SQL_ATTACH_TIME_RANGE_PARTITION,
    SQL_CREATE_TIME_RANGE_PARTITION,
    SQL_DETACH_PARTITION,
    SQL_SET_TABLE_TABLESPACE,
    PeriodType,
)
from .shortcuts import double_quote, drop_table, execute_sql, generate_set_indexes_tablespace_sql, single_quote


class PartitionConfig(models.Model):
    """You can get the configuration object of the partition table through ``Model.partitioning.config``,
    You can only edit the following fields via the object's ``save`` method:
    """

    model_label = models.TextField(unique=True)
    period = models.TextField(default=PeriodType.Month)
    """Partition period. you can only set options in the `PeriodType`.
    The default value is ``PeriodType.Month``. Changing this value will trigger the ``detach_partition`` method."""
    interval = models.PositiveIntegerField(null=True)
    """Detaching period. The ``detach_partition`` method defaults to detach partitions before the interval * period.
    The default is None, ie no partition will be detached. Changing this value will trigger the ``detach_partition`` method."""
    attach_tablespace = models.TextField(null=True)
    """The name of the tablespace specified when creating or attaching a partition. Modifying this field will only affect subsequent operations.
    A table migration may occur at this time."""
    detach_tablespace = models.TextField(null=True)
    """The name of the tablespace specified when detaching a partition. Modifying this field will only affect subsequent operations.
    A table migration may occur at this time."""

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        """This setting will take effect immediately when you modify the value of
        ``interval`` in the configuration.
        """

        adding = self._state.adding
        model = apps.get_model(self.model_label)

        if not adding:
            prev = self.__class__.objects.get(pk=self.pk)

        with transaction.atomic():
            super().save(force_insert, force_update, using, update_fields)
            if adding:
                # Creating first partition.
                model.partitioning.create_partition(0)

        if not adding:
            # Period or interval changed.
            if prev.period != self.period or (prev.interval != self.interval):
                model.partitioning.detach_partition()


class PartitionLog(models.Model):
    """You can only edit the following fields via the object's ``save`` method:"""

    config = models.ForeignKey(PartitionConfig, on_delete=models.CASCADE, related_name="logs")
    table_name = models.TextField(unique=True)
    # range bound: [start, end)
    start = models.DateTimeField()
    end = models.DateTimeField()
    is_attached = models.BooleanField(default=True)
    """Whether the partition is a attached partition. changing the value will trigger an attaching or detaching operation."""
    detach_time = models.DateTimeField(null=True)
    """When the value is not `None`, the partition will not be automatically detached before this time. The default is `None`."""

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        """This setting will take effect immediately when you modify the value of
        ``is_attached`` in the configuration.
        """

        model = apps.get_model(self.config.model_label)
        if self._state.adding:
            sql_sequence = [
                SQL_CREATE_TIME_RANGE_PARTITION
                % {
                    "parent": double_quote(model._meta.db_table),
                    "child": double_quote(self.table_name),
                    "date_start": single_quote(self.start.isoformat()),
                    "date_end": single_quote(self.end.isoformat()),
                }
            ]

            if self.config.attach_tablespace:
                sql_sequence[0] += SQL_APPEND_TABLESPACE % {"tablespace": self.config.attach_tablespace}
                sql_sequence.extend(generate_set_indexes_tablespace_sql(self.table_name, self.config.attach_tablespace))

            with transaction.atomic():
                super().save(force_insert, force_update, using, update_fields)
                execute_sql(sql_sequence)
                post_create_partition.send(sender=model, partition_log=self)
        else:
            with transaction.atomic():
                prev = self.__class__.objects.select_for_update().get(pk=self.pk)
                # Detach partition.
                if prev.is_attached and (not self.is_attached):
                    sql_sequence = [SQL_DETACH_PARTITION % {"parent": double_quote(model._meta.db_table), "child": double_quote(self.table_name)}]
                    if self.config.detach_tablespace:
                        sql_sequence.append(SQL_SET_TABLE_TABLESPACE % {"name": double_quote(self.table_name), "tablespace": self.config.detach_tablespace})
                        sql_sequence.extend(generate_set_indexes_tablespace_sql(self.table_name, self.config.detach_tablespace))

                    super().save(force_insert, force_update, using, update_fields)
                    execute_sql(sql_sequence)
                    post_detach_partition.send(sender=model, partition_log=self)
                # Attach partition.
                elif (not prev.is_attached) and self.is_attached:
                    sql_sequence = list()
                    if self.config.attach_tablespace:
                        sql_sequence.append(SQL_SET_TABLE_TABLESPACE % {"name": double_quote(self.table_name), "tablespace": self.config.attach_tablespace})
                        sql_sequence.extend(generate_set_indexes_tablespace_sql(self.table_name, self.config.attach_tablespace))

                    sql_sequence.append(
                        SQL_ATTACH_TIME_RANGE_PARTITION
                        % {
                            "parent": double_quote(model._meta.db_table),
                            "child": double_quote(self.table_name),
                            "date_start": single_quote(self.start.isoformat()),
                            "date_end": single_quote(self.end.isoformat()),
                        }
                    )

                    super().save(force_insert, force_update, using, update_fields)
                    execute_sql(sql_sequence)
                    post_attach_partition.send(sender=model, partition_log=self)
                # State has not changed.
                else:
                    super().save(force_insert, force_update, using, update_fields)

    @transaction.atomic
    def delete(self, using=None, keep_parents=False):
        """When the instance is deleted, the partition corresponding to it will also be deleted."""

        drop_table(self.table_name)
        super().delete(using, keep_parents)

    class Meta:
        ordering = ("-id",)



import logging
from importlib import import_module

from django.apps.config import MODELS_MODULE_NAME
from django.db.backends.postgresql.schema import DatabaseSchemaEditor
from pg_partitioning.manager import _PartitionManagerBase

logger = logging.getLogger(__name__)


default_create_model_method = DatabaseSchemaEditor.create_model
default_sql_create_table = DatabaseSchemaEditor.sql_create_table


def create_model(self, model):
    meta = model._meta
    try:
        cls_module = import_module(f"{meta.app_label}.{MODELS_MODULE_NAME}")
    except ModuleNotFoundError:
        cls_module = None
    cls = getattr(cls_module, meta.object_name, None)
    partitioning = getattr(cls, "partitioning", None)
    if isinstance(partitioning, _PartitionManagerBase):
        # XXX: Monkeypatch create_model.
        logger.debug("Partitioned model detected: %s", meta.label)
        _type = partitioning.type
        key = partitioning.partition_key
        DatabaseSchemaEditor.sql_create_table = f"CREATE TABLE %(table)s (%(definition)s) PARTITION BY {_type} ({key})"
        if meta.pk.name != key:
            """The partition key must be part of the primary key,
            and currently Django does not support setting a composite primary key,
            so its properties are turned off."""
            meta.pk.primary_key = False
            logger.info("Note that PK constraints for %s has been temporarily closed.", meta.label)
    else:
        DatabaseSchemaEditor.sql_create_table = default_sql_create_table
    default_create_model_method(self, model)
    meta.pk.primary_key = True


DatabaseSchemaEditor.create_model = create_model



from django.dispatch import Signal

post_create_partition = Signal(providing_args=["partition_log"])
"""Sent when a partition is created.
"""
post_attach_partition = Signal(providing_args=["partition_log"])
"""Sent when a partition is attached.
"""
post_detach_partition = Signal(providing_args=["partition_log"])
"""Sent when a partition is detached.
"""



"""
A Django extension that supports PostgreSQL 11 time ranges and list partitioning.
"""
REQUIRED_DJANGO_VERSION = [(2, 0), (2, 1)]
DJANGO_VERSION_ERROR = "django-pg-partitioning isn't available on the currently installed version of Django."

try:
    import django
except ImportError:
    raise ImportError(DJANGO_VERSION_ERROR)

if REQUIRED_DJANGO_VERSION[0] > tuple(django.VERSION[:2]) or tuple(django.VERSION[:2]) > REQUIRED_DJANGO_VERSION[1]:
    raise ImportError(DJANGO_VERSION_ERROR)


__version__ = "0.10dev1"

default_app_config = "pg_partitioning.apps.AppConfig"



import logging
from typing import Type

from django.db import models
from pg_partitioning.manager import ListPartitionManager, TimeRangePartitionManager

logger = logging.getLogger(__name__)


class _PartitioningBase:
    def __init__(self, partition_key: str, **options):
        self.partition_key = partition_key
        self.options = options

    def __call__(self, model: Type[models.Model]):
        if model._meta.abstract:
            raise NotImplementedError("Decorative abstract model classes are not supported.")


class TimeRangePartitioning(_PartitioningBase):
    """Use this decorator to declare the database table corresponding to the model to be partitioned by time range.

    Parameters:
      partition_key(str): Partition field name of DateTimeField.
      options: Currently supports the following keyword parameters:

        - default_period(PeriodType): Default partition period.
        - default_interval(int): Default detach partition interval.
        - default_attach_tablespace(str): Default tablespace for attached tables.
        - default_detach_tablespace(str): Default tablespace for attached tables.

    Example:
      .. code-block:: python

          from django.db import models
          from django.utils import timezone

          from pg_partitioning.manager import TimeRangePartitioning


          @TimeRangePartitioning(partition_key="timestamp")
          class MyLog(models.Model):
              name = models.TextField(default="Hello World!")
              timestamp = models.DateTimeField(default=timezone.now, primary_key=True)
    """

    def __call__(self, model: Type[models.Model]):
        super().__call__(model)
        if model._meta.get_field(self.partition_key).get_internal_type() != models.DateTimeField().get_internal_type():
            raise ValueError("The partition_key must be DateTimeField type.")
        model.partitioning = TimeRangePartitionManager(model, self.partition_key, self.options)
        return model


class ListPartitioning(_PartitioningBase):
    """Use this decorator to declare the database table corresponding to the model to be partitioned by list.

    Parameters:
      partition_key(str): Partition key name, the type of the key must be one of boolean, text or integer.

    Example:
      .. code-block:: python

          from django.db import models
          from django.utils import timezone

          from pg_partitioning.manager import ListPartitioning


          @ListPartitioning(partition_key="category")
          class MyLog(models.Model):
              category = models.TextField(default="A")
              timestamp = models.DateTimeField(default=timezone.now, primary_key=True)
    """

    def __call__(self, model: Type[models.Model]):
        super().__call__(model)
        support_field_types = [item.get_internal_type() for item in [models.TextField(), models.BooleanField(), models.IntegerField()]]
        if model._meta.get_field(self.partition_key).get_internal_type() not in support_field_types:
            raise NotImplementedError("The partition_key does not support this field type.")
        model.partitioning = ListPartitionManager(model, self.partition_key, self.options)
        return model



# Generated by Django 2.1.7 on 2019-02-17 12:00

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='PartitionConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('model_label', models.TextField(unique=True)),
                ('period', models.TextField(default='Month')),
                ('interval', models.PositiveIntegerField(null=True)),
                ('attach_tablespace', models.TextField(null=True)),
                ('detach_tablespace', models.TextField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name='PartitionLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('table_name', models.TextField(unique=True)),
                ('start', models.DateTimeField()),
                ('end', models.DateTimeField()),
                ('is_attached', models.BooleanField(default=True)),
                ('detach_time', models.DateTimeField(null=True)),
                ('config', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='logs', to='pg_partitioning.PartitionConfig')),
            ],
            options={
                'ordering': ('-id',),
            },
        ),
    ]



import logging
from typing import List, Optional, Tuple, Union

from django.db import connection
from pg_partitioning.constants import SQL_DROP_TABLE, SQL_GET_TABLE_INDEXES, SQL_SET_INDEX_TABLESPACE, SQL_SET_TABLE_TABLESPACE, SQL_TRUNCATE_TABLE

logger = logging.getLogger(__name__)


def single_quote(name: str) -> str:
    """Represent a string constants in SQL."""

    if name.startswith("'") and name.endswith("'"):
        return name
    return "'%s'" % name


def double_quote(name: str) -> str:
    """Represent a identify in SQL."""

    if name.startswith('"') and name.endswith('"'):
        return name
    return '"%s"' % name


def execute_sql(sql_sequence: Union[str, List[str], Tuple[str]], fetch: bool = False) -> Optional[Tuple]:
    """Execute SQL sequence and returning result."""

    sql_str = ""
    for statement in sql_sequence if isinstance(sql_sequence, (list, tuple)) else [sql_sequence]:
        sql_str += ";\n" + statement if sql_str else statement
    logger.debug("The sequence of SQL statements to be executed:\n %s", sql_str)
    with connection.cursor() as cursor:
        cursor.execute(sql_str)
        if fetch:
            return cursor.fetchall()


def generate_set_indexes_tablespace_sql(table_name: str, tablespace: str) -> List[str]:
    """Generate set indexes tablespace SQL sequence.

    Parameters:
      table_name(str): Table name.
      tablespace(str): Partition tablespace.
    """

    sql_sequence = []
    result = execute_sql(SQL_GET_TABLE_INDEXES % {"table_name": single_quote(table_name)}, fetch=True)
    for item in result:
        sql_sequence.append(SQL_SET_INDEX_TABLESPACE % {"name": double_quote(item[0]), "tablespace": tablespace})
    return sql_sequence


def set_tablespace(table_name: str, tablespace: str) -> None:
    """Set the tablespace for a table and indexes.

    Parameters:
      table_name(str): Table name.
      tablespace(str): Tablespace name.
    """

    sql_sequence = [SQL_SET_TABLE_TABLESPACE % {"name": double_quote(table_name), "tablespace": tablespace}]
    sql_sequence.extend(generate_set_indexes_tablespace_sql(table_name, tablespace))
    execute_sql(sql_sequence)


def truncate_table(table_name: str) -> None:
    """Truncate table.

    Parameters:
      table_name(str): Table name.
    """

    execute_sql(SQL_TRUNCATE_TABLE % {"name": double_quote(table_name)})


def drop_table(table_name: str) -> None:
    """Drop table.

    Parameters:
      table_name(str): Table name.
    """

    execute_sql(SQL_DROP_TABLE % {"name": double_quote(table_name)})



import os

from setuptools import setup


def rel(*xs):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), *xs)


with open(rel("README.rst")) as f:
    long_description = f.read()


with open(rel("pg_partitioning", "__init__.py"), "r") as f:
    version_marker = "__version__ = "
    for line in f:
        if line.startswith(version_marker):
            _, version = line.split(version_marker)
            version = version.strip().strip('"')
            break
    else:
        raise RuntimeError("Version marker not found.")


dependencies = [
    "python-dateutil~=2.7",
]

extra_dependencies = {
    "django": [
        "Django>=2.0,<2.2"
    ],
}

extra_dependencies["all"] = list(set(sum(extra_dependencies.values(), [])))
extra_dependencies["dev"] = extra_dependencies["all"] + [
    # Pinned due to https://bitbucket.org/ned/coveragepy/issues/578/incomplete-file-path-in-xml-report
    "coverage>=4.0,<4.4",

    # Docs
    "alabaster==0.7.12",
    "sphinx==1.8.3",
    "sphinxcontrib-napoleon==0.7",

    # Linting
    "flake8~=3.6.0",
    "isort~=4.3.4",
    "black~=18.9b0",
    "flake8-bugbear~=18.8.0",
    "flake8-quotes~=1.0.0",

    # Misc
    "dj-database-url==0.5.0",
    "psycopg2-binary==2.7.6.1",
    "twine==1.12.1",

    # Testing
    "tox==3.6.1",
    "tox-venv"
]


setup(
    name="django-pg-partitioning",
    version=version,
    author="Boyce Li",
    author_email="monobiao@gmail.com",
    description="A Django extension that supports PostgreSQL 11 time ranges and list partitioning.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/chaitin/django-pg-partitioning",
    packages=["pg_partitioning", "pg_partitioning.migrations", "pg_partitioning.patch"],
    include_package_data=True,
    install_requires=dependencies,
    extras_require=extra_dependencies,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
)



import datetime
from collections import Iterable
from typing import Optional, Type, Union

import pytz
from dateutil.relativedelta import MO, relativedelta
from django.conf import settings
from django.db import IntegrityError, models
from django.db.models import Q
from django.utils import timezone
from pg_partitioning.shortcuts import double_quote, execute_sql, generate_set_indexes_tablespace_sql, single_quote

from .constants import (
    DT_FORMAT,
    SQL_APPEND_TABLESPACE,
    SQL_ATTACH_LIST_PARTITION,
    SQL_CREATE_LIST_PARTITION,
    SQL_DETACH_PARTITION,
    SQL_SET_TABLE_TABLESPACE,
    PartitioningType,
    PeriodType,
)
from .models import PartitionConfig, PartitionLog


class _PartitionManagerBase:
    type = None

    def __init__(self, model: Type[models.Model], partition_key: str, options: dict):
        self.model = model
        self.partition_key = partition_key
        self.options = options


class TimeRangePartitionManager(_PartitionManagerBase):
    """Manage time-based partition APIs."""

    type = PartitioningType.Range

    @property
    def config(self) -> PartitionConfig:
        """Get the latest PartitionConfig instance of this model.
        In order to avoid the race condition, we used **select_for_update** when querying.

        Returns:
          PartitionConfig: The latest PartitionConfig instance of this model.
        """
        try:
            return PartitionConfig.objects.select_for_update().get(model_label=self.model._meta.label_lower)
        except PartitionConfig.DoesNotExist:
            try:
                return PartitionConfig.objects.create(
                    model_label=self.model._meta.label_lower,
                    period=self.options.get("default_period", PeriodType.Month),
                    interval=self.options.get("default_interval"),
                    attach_tablespace=self.options.get("default_attach_tablespace"),
                    detach_tablespace=self.options.get("default_detach_tablespace"),
                )
            except IntegrityError:
                return PartitionConfig.objects.select_for_update().get(model_label=self.model._meta.label_lower)

    @property
    def latest(self) -> Optional[PartitionLog]:
        """Get the latest PartitionLog instance of this model.

        Returns:
          Optional[PartitionLog]: The latest PartitionLog instance of this model or none.
        """
        return self.config.logs.order_by("-id").first()

    @classmethod
    def _get_period_bound(cls, date_start, initial, addition_zeros=None, is_week=False, **kwargs):
        zeros = {"hour": 0, "minute": 0, "second": 0, "microsecond": 0}
        if addition_zeros:
            zeros.update(addition_zeros)

        def func():  # lazy evaluation
            if initial:
                start = date_start.replace(**zeros)
                if is_week:
                    start -= relativedelta(days=start.weekday())
            else:
                start = date_start
            end = start + relativedelta(**kwargs, **zeros)
            return start, end

        return func

    def create_partition(self, max_days_to_next_partition: int = 1) -> None:
        """The partition of the next cycle is created according to the configuration.
        After modifying the period field, the new period will take effect the next time.
        The start time of the new partition is the end time of the previous partition table,
        or the start time of the current archive period when no partition exists.

        For example:
        the current time is June 5, 2018, and the archiving period is one year,
        then the start time of the first partition is 00:00:00 on January 1, 2018.

        Parameters:
          max_days_to_next_partition(int):
            If numbers of days remained in current partition is greater than ``max_days_to_next_partition``, no new partitions will be created.
        """
        while True:
            if max_days_to_next_partition > 0 and self.latest and timezone.now() < (self.latest.end - relativedelta(days=max_days_to_next_partition)):
                return

            partition_timezone = getattr(settings, "PARTITION_TIMEZONE", None)
            if partition_timezone:
                partition_timezone = pytz.timezone(partition_timezone)
            date_start = timezone.localtime(self.latest.end if self.latest else None, timezone=partition_timezone)
            initial = not bool(self.latest)
            date_start, date_end = {
                PeriodType.Day: self._get_period_bound(date_start, initial, days=+1),
                PeriodType.Week: self._get_period_bound(date_start, initial, is_week=True, days=+1, weekday=MO),
                PeriodType.Month: self._get_period_bound(date_start, initial, addition_zeros=dict(day=1), months=+1),
                PeriodType.Year: self._get_period_bound(date_start, initial, addition_zeros=dict(month=1, day=1), years=+1),
            }[self.config.period]()

            partition_table_name = "_".join((self.model._meta.db_table, date_start.strftime(DT_FORMAT), date_end.strftime(DT_FORMAT)))
            PartitionLog.objects.create(config=self.config, table_name=partition_table_name, start=date_start, end=date_end)

            if not max_days_to_next_partition > 0:
                return

    def attach_partition(self, partition_log: Optional[Iterable] = None, detach_time: Optional[datetime.datetime] = None) -> None:
        """Attach partitions.

        Parameters:
          partition_log(Optional[Iterable]):
            All partitions are attached when you don't specify partitions to attach.
          detach_time(Optional[datetime.datetime]):
            When the partition specifies the archive time, it will **not** be automatically archived until that time.
        """
        if not partition_log:
            partition_log = PartitionLog.objects.filter(config=self.config, is_attached=False)

        for log in partition_log:
            log.is_attached = True
            log.detach_time = detach_time
            log.save()

    def detach_partition(self, partition_log: Optional[Iterable] = None) -> None:
        """Detach partitions.

        Parameters:
          partition_log(Optional[Iterable]):
            Specify a partition to archive. When you don't specify a partition to archive, all partitions that meet the configuration rule are archived.
        """
        if not partition_log:
            if self.config.interval:
                # fmt: off
                period = {PeriodType.Day: {"days": 1},
                          PeriodType.Week: {"weeks": 1},
                          PeriodType.Month: {"months": 1},
                          PeriodType.Year: {"years": 1}}[self.config.period]
                # fmt: on
                now = timezone.now()
                detach_timeline = now - self.config.interval * relativedelta(**period)
                partition_log = PartitionLog.objects.filter(config=self.config, end__lt=detach_timeline, is_attached=True)
                partition_log = partition_log.filter(Q(detach_time=None) | Q(detach_time__lt=now))
            else:
                return

        for log in partition_log:
            log.is_attached = False
            log.detach_time = None
            log.save()

    def delete_partition(self, partition_log: Iterable) -> None:
        """Delete partitions.

        Parameters:
          partition_log(Iterable): The partitions to be deleted.
        """
        for log in partition_log:
            if log.config == self.config:
                log.delete()


def _db_value(value: Union[str, int, bool, None]) -> str:
    if value is None:
        return "null"
    return single_quote(value) if isinstance(value, str) else str(value)


class ListPartitionManager(_PartitionManagerBase):
    """Manage list-based partition APIs."""

    type = PartitioningType.List

    def create_partition(self, partition_name: str, value: Union[str, int, bool, None], tablespace: str = None) -> None:
        """Create partitions.

        Parameters:
          partition_name(str): Partition name.
          value(Union[str, int, bool, None]): Partition key value.
          tablespace(str): Partition tablespace name.
        """

        sql_sequence = [
            SQL_CREATE_LIST_PARTITION % {"parent": double_quote(self.model._meta.db_table), "child": double_quote(partition_name), "value": _db_value(value)}
        ]
        if tablespace:
            sql_sequence[0] += SQL_APPEND_TABLESPACE % {"tablespace": tablespace}
            sql_sequence.extend(generate_set_indexes_tablespace_sql(partition_name, tablespace))
        execute_sql(sql_sequence)

    def attach_partition(self, partition_name: str, value: Union[str, int, bool, None], tablespace: str = None) -> None:
        """Attach partitions.

        Parameters:
          partition_name(str): Partition name.
          value(Union[str, int, bool, None]): Partition key value.
          tablespace(str): Partition tablespace name.
        """

        sql_sequence = list()
        if tablespace:
            sql_sequence.append(SQL_SET_TABLE_TABLESPACE % {"name": double_quote(partition_name), "tablespace": tablespace})
            sql_sequence.extend(generate_set_indexes_tablespace_sql(partition_name, tablespace))
        sql_sequence.append(
            SQL_ATTACH_LIST_PARTITION
            % {"parent": double_quote(self.model._meta.db_table), "child": double_quote(partition_name), "value": single_quote(_db_value(value))}
        )
        execute_sql(sql_sequence)

    def detach_partition(self, partition_name: str, tablespace: str = None) -> None:
        """Detach partitions.

        Parameters:
          partition_name(str): Partition name.
          tablespace(str): Partition tablespace name.
        """
        sql_sequence = [SQL_DETACH_PARTITION % {"parent": double_quote(self.model._meta.db_table), "child": double_quote(partition_name)}]
        if tablespace:
            sql_sequence.append(SQL_SET_TABLE_TABLESPACE % {"name": double_quote(partition_name), "tablespace": tablespace})
            sql_sequence.extend(generate_set_indexes_tablespace_sql(partition_name, tablespace))
        execute_sql(sql_sequence)



SQL_CREATE_TIME_RANGE_PARTITION = """\
CREATE TABLE IF NOT EXISTS %(child)s PARTITION OF %(parent)s FOR VALUES FROM (%(date_start)s) TO (%(date_end)s)"""
SQL_CREATE_LIST_PARTITION = """\
CREATE TABLE IF NOT EXISTS %(child)s PARTITION OF %(parent)s FOR VALUES IN (%(value)s)"""
SQL_SET_TABLE_TABLESPACE = """\
ALTER TABLE IF EXISTS %(name)s SET TABLESPACE %(tablespace)s"""
SQL_APPEND_TABLESPACE = " TABLESPACE %(tablespace)s"
SQL_ATTACH_TIME_RANGE_PARTITION = """\
ALTER TABLE IF EXISTS %(parent)s ATTACH PARTITION %(child)s FOR VALUES FROM (%(date_start)s) TO (%(date_end)s)"""
SQL_ATTACH_LIST_PARTITION = """\
ALTER TABLE IF EXISTS %(parent)s ATTACH PARTITION %(child)s FOR VALUES IN (%(value)s)"""
SQL_DETACH_PARTITION = "ALTER TABLE IF EXISTS %(parent)s DETACH PARTITION %(child)s"
SQL_DROP_TABLE = "DROP TABLE IF EXISTS %(name)s"
SQL_TRUNCATE_TABLE = "TRUNCATE TABLE %(name)s"
SQL_DROP_INDEX = "DROP INDEX IF EXISTS %(name)s"
SQL_CREATE_INDEX = "CREATE INDEX IF NOT EXISTS %(name)s ON %(table_name)s USING %(method)s (%(column_name)s)"
SQL_SET_INDEX_TABLESPACE = "ALTER INDEX %(name)s SET TABLESPACE %(tablespace)s"
SQL_GET_TABLE_INDEXES = "SELECT indexname FROM pg_indexes WHERE tablename = %(table_name)s"

DT_FORMAT = "%Y-%m-%d"


class PartitioningType:
    Range = "RANGE"
    List = "LIST"


class PeriodType:
    Day = "Day"
    Week = "Week"
    Month = "Month"
    Year = "Year"
